%% Emacs, this is -*-prolog-*-

:- dynamic aspect/2.
:- dynamic process/2.

:- consult('xml-facts-mapping').
:- consult('xml-facts-parse').
:- consult('xml-facts-generate').

:- consult('aspect-facts-generate').
:- consult('aspect-to-process').
:- consult('pointcuts').
:- consult('advisor').
:- consult('run').
:- consult('strings').
:- consult('logging').

test1b :-
        clear,
        defineAspects(['Logging'],['/windows/D/workspace/padus/demo/Logging.xml']),
        defineProcesses(['ConferenceCall'],['/windows/D/workspace/padus/demo/ConferenceCall.bpel']),
    %    setOutputDir('/windows/D/workspace/padus/demo/'),
        deploy('/windows/D/workspace/padus/demo/deployment.xml').

test1c :-
        clear,
        defineAspects(['Logging','DLogging'],['/windows/D/workspace/padus/demo/Logging.xml','/windows/D/workspace/padus/demo/DLogging.xml']),
        defineProcesses(['ConferenceCall','Demonstrator'],['/windows/D/workspace/padus/demo/ConferenceCall.bpel','/windows/D/workspace/padus/demo/Demonstrator.bpel']),
        setOutputDir('/windows/D/workspace/padus/demo/'),
        deploy('/windows/D/workspace/padus/demo/deployment_3.xml').

%%  INITIALISATION
%% ----------------

clear :- clear(_).
clear(Type) :- clear_base(Type), clear_advice(Type), clear_pointcuts, clear_advices.
clearOutputDir :- retractall(outputDir(_)).

%% Associate an aspect with an xml file 
defineAspects(Aliases,Files) :-
        assert_aspects(Aliases,Files).

assert_aspects([],[]).
assert_aspects([Alias|Aliases],[File|Files]):-
        assert(aspect(Alias, File)),
        myprint( 'Aspect Defenitions: ', [Alias|Aliases], [File|Files], 'DEPLOYMENT'),
        flatten(Aliases, Aa),
        flatten(Files, Bb),
        assert_aspects(Aa, Bb).
%assert_aspects(_,_).



%% Associate a process with an xml file
defineProcesses(Aliases,Files) :-
        assert_processes(Aliases,Files).

assert_processes([],[]).
assert_processes([Alias|Aliases],[File|Files]):-
        assert(process(Alias,File)),
        myprint('Process Defenitions: ',[Alias|Aliases],[File|Files], 'DEPLOYMENT'),
        flatten(Aliases,Aa),
        flatten(Files,Bb),
        assert_processes(Aa,Bb).
%assert_processes(_,_).

setOutputDir(Dir) :-
        assert(outputDir(Dir)).

getFile(Value, DestFile) :-
		(clause(outputDir(D), Head) -> outputDir(Dir); =(Dir, '')),
		concat(Dir, Value, R1),
		concat(R1, '$weaved.bpel', DestFile).



%%  Deployment
%% ------------

deploy(File) :-
        myprint('*********************', 'INFO'), myprint('* Parse Deployment: *', 'INFO'), myprint('*********************', 'INFO'),
        loadAndParseFile(File,'deployment'),
        generateAndPrint('deployment'),
        myprint('Iterating over Processes... ', 'INFO'),
        iterateOverProcesses.

resolveAspectName(AspectNameValue, ProcessValue, IdValue) :-
        findall(Id1, (attribute(Id1,'process', ProcessValue, 'deployment'),
                      attribute(Id2,'id', IdValue, 'deployment'),
                      Id1=Id2,
                      node(Id1, NSAspect, 'deployment'),
                      split(NSAspect, _, ':','aspect')), [AspectNodeId]),
        attribute(AspectNodeId, 'name', AspectNameValue, 'deployment').



%%   PROCESSES
%% -------------

iterateOverProcesses :-
		debug('Iterating over processes...'),
		findall(PrecedenceNodeId,
		        (child(0, PrecedenceNodeId, _, 'deployment'),
		         node(PrecedenceNodeId, PrecedenceNodeName, 'deployment'),
		         split(PrecedenceNodeName, _, ':', 'precedence')),
		        PrecedenceNodeIds),
		debug(['PrecedenceNodeIds = ', PrecedenceNodeIds]),
		loadProcessInStructure(PrecedenceNodeIds),
		clearOutputDir.

%% Get process file from deployment structure and parse to facts
loadProcessInStructure([PrecedenceNodeId | PrecedenceNodeIds]) :-
		info(['Started weaving...']),
        attribute(PrecedenceNodeId, 'process', ProcessName, 'deployment'),
        process(ProcessName, File),
        debug(['Process \'', ProcessName, '\' is specified in \'', File, '\'']),
		debug(''),
		debug('**************************************************************************'),
		debug(['* Parsing process \'', ProcessName, '\'...']),
		debug('**************************************************************************'),
		debug(''),
		loadAndParseFile(File, 'process'),
%		generateAndPrint('process'),
        iterateOverAspects(PrecedenceNodeId),
		debug(''),
		debug('**************************************************************************'),
		debug('* Generating resulting process...'),
		debug('**************************************************************************'),
		debug(''),
%		generateAndPrint('process'),
		getFile(ProcessName, DestFile),
		generateAndWrite(DestFile),
		info(['Finished weaving; generated \'', DestFile, '\'.']),
        clear_base('process'),
        loadProcessInStructure(PrecedenceNodeIds).
loadProcessInStructure([]).



%%   ASPECTS
%% -----------
iterateOverAspects(PrecedenceNodeId) :-
		debug('Iterating over aspects...'),
		findall(AspectNodeId,
		        (child(PrecedenceNodeId, AspectNodeId, _, 'deployment'),
		         node(AspectNodeId, AspectNodeName, 'deployment'),
		         split(AspectNodeName, _, ':', 'aspect')),
		        AspectNodeIds),
        debug(['AspectNodeIds = ', AspectNodeIds]),
        loadAspectInStructure(AspectNodeIds, PrecedenceNodeId).

%% Get aspect file from deployment structure
loadAspectInStructure([AspectNodeId | AspectNodeIds], PrecedenceNodeId) :-
 		attribute(AspectNodeId, 'id', AspectId, 'deployment'),
        attribute(PrecedenceNodeId, 'process', ProcessName, 'deployment'),
        resolveAspectName(AspectName, ProcessName, AspectId),
        aspect(AspectName, File),
        debug(['Aspect \'', AspectName, '\' is specified in \'', File, '\'']),
		debug(''),
		debug('**************************************************************************'),
		debug(['* Parsing aspect \'', AspectName, '\'...']),
		debug('**************************************************************************'),
		debug(''),
        loadAndParseFile(File, 'aspect'),
%		generateAndPrint('aspect'),
        iterateOverIncludes,
        iterateOverAdviceConstructs,
        weaveAspectsInProcess,
        weaveUsingStatements,
%		myprint('***************************', 'INFO'),
%		myprint('* Resulting AspectID: ',AspectNodeId,'*','INFO'),
%		myprint('***************************', 'INFO'),
%		generateAndPrint('aspect'),
        clear_base('aspect'),
        loadAspectInStructure(AspectNodeIds, PrecedenceNodeId).
loadAspectInStructure([], _).



%%% LOAD & PARSE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Load and parse file

loadAndParseBpelProcess(File) :-
	loadAndParseFile(File, 'process').

loadAndParseFile(File, RootTag) :-
	loadAndParseFile(File, RootTag, NewStructure).

loadAndParseFile(File, RootTag, NewStructure) :-
	debug(['Loading and parsing \'', File, '\'...']),
	loadFileInStructure(File, Structure),
	loadStructureWithRootTag(RootTag, Structure, NewStructure),
	parse(NewStructure, 0, RootTag),
	!.

% Load and parse next file

% TODO
%  - NextId specific for space

loadAndParseNextFile(File, NextId):-
        loadAndParseNextFile(File, NextId, 'disable').
                
loadAndParseNextFile(File, NextId, RootTag):-
        loadAndParseNextFile(File, NextId, RootTag, NewStructure).

loadAndParseNextFile(File, NextId, RootTag, NewStructure):-
        loadFileInStructure(File, Structure),    
        loadStructureWithRootTag( RootTag, Structure, NewStructure),
        parseNextStructure(NewStructure, NextId, RootTag).

parseNextStructure(Structure, NextId, RootTag):-
        maxId(Id, RootTag),
        next(Id, NextId),
        parse(Structure, NextId, RootTag).





% Load Structure according to RootTag

loadStructureWithRootTag(_, [], NewStructure) :-
		Structure = [].

loadStructureWithRootTag( RootTag, Structure, NewStructure):-
        RootTag = 'disable', !, NewStructure = Structure.

loadStructureWithRootTag( RootTag, [element( NS, _Atts, Children) | Siblings], NewStructure):-  
        myprint('namespace: ', NS, 'DEPLOYMENT'),
        split( NS, _, ':', RootTag), !,  myprint('found!!!!!!', 'DEPLOYMENT'),
        =(NewStructure, [element( NS, _Atts, Children) | Siblings]).

loadStructureWithRootTag( RootTag, [element( NS, _Atts, Children) | Siblings], NewStructure):-
        %myprint('children: ',Children, 'DEPLOYMENT'),
        %myprint('siblings: ',Siblings, 'DEPLOYMENT'),
        loadStructureWithRootTag( RootTag, Children, NewStructure),
        loadStructureWithRootTag( RootTag, Siblings, NewStructure).

%% Use Prolog library to parse xml-files
loadFileInStructure(File, Structure) :-
        load_structure(File, Structure, [dialect(xml), space(remove)]).

%% Split namespace from actual tag
split(ABC, A, B, C) :-
        concat(A, BC, ABC),
        concat(B, C, BC).

% If no namespace is used
split(ABC, _, ':', ABC) :- not(sub_string(ABC,A,B,C,':')).


myprint(Aa,Bb,Cc,Dd, Type) :-
        logging(STypes),
        member(Type, STypes),!,
        write(Aa),
        write(Bb),
        write(' / '),
        write(Cc),
        write(' / '),
        write_ln(Dd).
myprint(Aa,Bb,Cc,Dd, Type).


myprint(Aa,Bb,Cc, Type) :-
        logging(STypes),
        member(Type, STypes),!,
        write(Aa),
        write(Bb),
        write(' / '),
        write_ln(Cc).
myprint(Aa,Bb,Cc, Type).

myprint(Aa,Bb, Type) :-
        logging(STypes),
        member(Type, STypes),!,
        write(Aa),
        write_ln(Bb).
myprint(Aa,Bb, Type).

myprint(Aa, Type) :-
        logging(STypes),
        member(Type, STypes),!,
        write_ln(Aa).
myprint(Aa, Type).



logging(Types):-
        Types = ['DEBUG', 'INFO'].

% 'PARSE','GENERATE','ASPECTS','ATOP','DEPLOYMENT'
