% Predicates for selecting joinpoints:

pc_(Jp, Activity, []) :-
	node(Jp, NSActivity, 'process'),
	split(NSActivity, _, ':', Activity),
	(maxProcessId(MaxId, 'process'), Jp < MaxId).

pc_(Jp, Activity, [Attr | Attributes]) :-
	Attr =.. [Key, Value],
	attribute(Jp, Key, NSValue, 'process'),
	split(NSValue, _, ':', Value),
	pc_(Jp, Activity, Attributes).

% Pointcut predicates:

invoking(Jp, Attrs) :-
	pc_(Jp, invoke, Attrs).
invoking(Jp) :-
	invoking(Jp, []).
invoking(Jp, PartnerLink, PortType, Operation) :-
	invoking(Jp, [partnerLink(PartnerLink), portType(PortType), operation(Operation)]).
invoking(Jp, Name, PartnerLink, PortType, Operation) :-
	invoking(Jp, [name(Name), partnerLink(PartnerLink), portType(PortType), operation(Operation)]).
invoking(Jp, Name, PartnerLink, PortType, Operation, InputVariable, OutputVariable) :-
	invoking(Jp, [name(Name), partnerLink(PartnerLink), portType(PortType), operation(Operation), inputVariable(InputVariable), outputVariable(OutputVariable)]).

receiving(Jp, Attrs) :- pc_(Jp, receive, Attrs).
receiving(Jp) :- receiving(Jp, []).

replying(Jp, Attrs) :- pc_(Jp, reply, Attrs).
replying(Jp) :- replying(Jp, []).

assigning(Jp, Attrs) :- pc_(Jp, assign, Attrs).
assigning(Jp) :- assigning(Jp, []).

scoping(Jp, Attrs) :-  pc_(Jp, scope, Attrs).
scoping(Jp) :- scoping(Jp, []).

picking(Jp, Attrs) :- pc_(Jp, pick, Attrs).
picking(Jp) :- picking(Jp, []).

switching(Jp, Attrs) :- pc_(Jp, switch, Attrs).
switching(Jp) :- switching(Jp, []).

flowing(Jp, Attrs) :- pc_(Jp, flow, Attrs).
flowing(Jp) :- flowing(Jp, []).

% ...

% Supplementary predicates:

% Links a joinpoint with the process it is defined in:
inProcess(Joinpoint, Attrs) :- pc_(Jp, process, Attrs).

% Links a joinpoint with the process instance it occurs in:
inProcessInstance(Joinpoint, ProcessInstance) :- write_ln('Not implemented yet!').

% Links the name of a variable to its value in a specific process instance:
variableValue(ProcessInstance, Name, Value) :- write_ln('Not implemented yet!').

variableValue(ProcessInstance, Name, Value, Advice, [element(switch, [], [element(case, [condition=Condition], [Advice])])]) :-
	concat('getVariableData(\'', Name, A),
	concat(A, '\')=\'', B),
	concat(B, Value, C),
	concat(C, '\'', Condition).
